@@ -52,7 +52,7 @@ import vobject
 
 from . import auth, rights, storage, web, xmlutils
 
-VERSION = "2.1.1"
+VERSION = "2.1.4"
 
 NOT_ALLOWED = (
     client.FORBIDDEN, (("Content-Type", "text/plain"),),
@@ -118,7 +118,7 @@ class HTTPServer(wsgiref.simple_server.WSGIServer):
                 self.max_connections)
         else:
             # use dummy context manager
-            self.connections_guard = contextlib.suppress()
+            self.connections_guard = contextlib.ExitStack()
 
     def get_request(self):
         # Set timeout for client
@@ -154,7 +154,8 @@ class HTTPSServer(HTTPServer):
             cert_reqs=ssl.CERT_REQUIRED if self.certificate_authority else
             ssl.CERT_NONE,
             ca_certs=self.certificate_authority or None,
-            ssl_version=self.protocol, ciphers=self.ciphers)
+            ssl_version=self.protocol, ciphers=self.ciphers,
+            do_handshake_on_connect=False)
 
         self.server_bind()
         self.server_activate()
@@ -168,6 +169,19 @@ class ThreadedHTTPServer(socketserver.ThreadingMixIn, HTTPServer):
 
 class ThreadedHTTPSServer(socketserver.ThreadingMixIn, HTTPSServer):
     def process_request_thread(self, request, client_address):
+        try:
+            try:
+                request.do_handshake()
+            except socket.timeout:
+                raise
+            except Exception as e:
+                raise RuntimeError("SSL handshake failed: %s" % e) from e
+        except Exception:
+            try:
+                self.handle_error(request, client_address)
+            finally:
+                self.shutdown_request(request)
+            return
         with self.connections_guard:
             return super().process_request_thread(request, client_address)
 
@@ -289,7 +303,7 @@ class Application:
                 write_allowed_items.append(item)
             self.logger.debug(
                 "%s has %s access to %s",
-                "%r" % user if user else "anonymous user",
+                repr(user) if user else "anonymous user",
                 " and ".join(text_status) if text_status else "NO", target)
         return read_allowed_items, write_allowed_items
 
@@ -355,7 +369,7 @@ class Application:
 
         remote_host = "unknown"
         if environ.get("REMOTE_HOST"):
-            remote_host = "%r" % environ["REMOTE_HOST"]
+            remote_host = repr(environ["REMOTE_HOST"])
         elif environ.get("REMOTE_ADDR"):
             remote_host = environ["REMOTE_ADDR"]
         if environ.get("HTTP_X_FORWARDED_FOR"):
@@ -444,7 +458,12 @@ class Application:
                         None)
                 if not principal:
                     with self.Collection.acquire_lock("w", user):
-                        self.Collection.create_collection(principal_path)
+                        try:
+                            self.Collection.create_collection(principal_path)
+                        except ValueError as e:
+                            self.logger.warning("Failed to create principal "
+                                                "collection %r: %s", user, e)
+                            is_authenticated = False
             else:
                 self.logger.warning("Access to principal path %r denied by "
                                     "rights backend", principal_path)
@@ -464,7 +483,7 @@ class Application:
                 environ, base_prefix, path, user)
             if (status, headers, answer) == NOT_ALLOWED:
                 self.logger.info("Access to %r denied for %s", path,
-                                 "%r" % user if user else "anonymous user")
+                                 repr(user) if user else "anonymous user")
         else:
             status, headers, answer = NOT_ALLOWED
 
@@ -621,6 +640,7 @@ class Application:
             # TODO: use this?
             # timezone = props.get("C:calendar-timezone")
             try:
+                storage.check_and_sanitize_props(props)
                 self.Collection.create_collection(path, props=props)
             except ValueError as e:
                 self.logger.warning(
@@ -647,6 +667,7 @@ class Application:
                 return WEBDAV_PRECONDITION_FAILED
             props = xmlutils.props_from_request(xml_content)
             try:
+                storage.check_and_sanitize_props(props)
                 self.Collection.create_collection(path, props=props)
             except ValueError as e:
                 self.logger.warning(
@@ -764,8 +785,13 @@ class Application:
                 return WEBDAV_PRECONDITION_FAILED
             headers = {"DAV": DAV_HEADERS,
                        "Content-Type": "text/xml; charset=%s" % self.encoding}
-            xml_answer = xmlutils.proppatch(base_prefix, path, xml_content,
-                                            item)
+            try:
+                xml_answer = xmlutils.proppatch(base_prefix, path, xml_content,
+                                                item)
+            except ValueError as e:
+                self.logger.warning(
+                    "Bad PROPPATCH request on %r: %s", path, e, exc_info=True)
+                return BAD_REQUEST
             return (client.MULTI_STATUS, headers,
                     self._write_xml_content(xml_answer))
 
@@ -814,36 +840,53 @@ class Application:
                 return PRECONDITION_FAILED
 
             try:
-                items = list(vobject.readComponents(content or ""))
+                items = tuple(vobject.readComponents(content or ""))
+                if not write_whole_collection and len(items) != 1:
+                    raise RuntimeError(
+                        "Item contains %d components" % len(items))
+                if write_whole_collection or not parent_item.get_meta("tag"):
+                    content_type = environ.get("CONTENT_TYPE",
+                                               "").split(";")[0]
+                    tags = {value: key
+                            for key, value in xmlutils.MIMETYPES.items()}
+                    tag = tags.get(content_type)
+                    if items and items[0].name == "VCALENDAR":
+                        tag = "VCALENDAR"
+                    elif items and items[0].name in ("VCARD", "VLIST"):
+                        tag = "VADDRESSBOOK"
+                else:
+                    tag = parent_item.get_meta("tag")
+                if tag == "VCALENDAR" and len(items) > 1:
+                    raise RuntimeError("VCALENDAR collection contains %d "
+                                       "components" % len(items))
                 for i in items:
                     storage.check_and_sanitize_item(
                         i, is_collection=write_whole_collection, uid=item.uid
-                        if not write_whole_collection and item else None)
+                        if not write_whole_collection and item else None,
+                        tag=tag)
             except Exception as e:
                 self.logger.warning(
                     "Bad PUT request on %r: %s", path, e, exc_info=True)
                 return BAD_REQUEST
-            content_type = environ.get("CONTENT_TYPE", "").split(";")[0]
-            tags = {value: key for key, value in xmlutils.MIMETYPES.items()}
-            tag = tags.get(content_type)
-            if items and items[0].name == "VCALENDAR":
-                tag = "VCALENDAR"
-            elif items and items[0].name == "VCARD":
-                tag = "VADDRESSBOOK"
 
             if write_whole_collection:
+                props = {"tag": tag} if tag else {}
                 try:
+                    storage.check_and_sanitize_props(props)
                     new_item = self.Collection.create_collection(
-                        path, items, {"tag": tag} if tag else None)
+                        path, items, props)
                 except ValueError as e:
                     self.logger.warning(
                         "Bad PUT request on %r: %s", path, e, exc_info=True)
                     return BAD_REQUEST
             else:
-                if tag and not parent_item.get_meta("tag"):
-                    parent_item.set_meta({"tag": tag})
                 href = posixpath.basename(path.strip("/"))
                 try:
+                    if tag and not parent_item.get_meta("tag"):
+                        new_props = parent_item.get_meta()
+                        new_props["tag"] = tag
+                        storage.check_and_sanitize_props(new_props)
+                        parent_item.set_meta_all(new_props)
                     new_item = parent_item.upload(href, items[0])
                 except ValueError as e:
                     self.logger.warning(
