@@ -66,10 +66,6 @@ TIMESTAMP_MIN = math.floor(DATETIME_MIN.timestamp())
 TIMESTAMP_MAX = math.ceil(DATETIME_MAX.timestamp())
 
 
-class VObjectBugException(Exception):
-    """Exception for workarounds related to bugs in VObject."""
-
-
 def pretty_xml(element, level=0):
     """Indent an ElementTree ``element`` and its children."""
     if not level:
@@ -250,12 +246,12 @@ def _time_range_match(vobject_item, filter_, child_name):
 
     matched = False
 
-    def range_fn(range_start, range_end):
+    def range_fn(range_start, range_end, recurrence):
         nonlocal matched
         if start < range_end and range_start < end:
             matched = True
             return True
-        if end < range_start:
+        if end < range_start and not recurrence:
             return True
         return False
 
@@ -271,8 +267,8 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
     `vobject_item`` with visitors ``range_fn`` and ``infinity_fn``.
 
     ``range_fn`` gets called for every time_range with ``start`` and ``end``
-    datetimes as arguments. If the function returns True, the operation is
-    cancelled.
+    datetimes and ``recurrence`` as arguments. If the function returns True,
+    the operation is cancelled.
 
     ``infinity_fn`` gets called when an infiite recurrence rule is detected
     with ``start`` datetime as argument. If the function returns True, the
@@ -281,41 +277,52 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
     See rfc4791-9.9.
 
     """
+
+    # HACK: According to rfc5545-3.8.4.4 an recurrance that is resheduled
+    # with Recurrence ID affects the recurrence itself and all following
+    # recurrences too. This is not respected and client don't seem to bother
+    # either.
+
     child = getattr(vobject_item, child_name.lower())
 
-    # TODO: Single recurrences can be overwritten by components with
-    # RECURRENCE-ID (http://www.kanzaki.com/docs/ical/recurrenceId.html). They
-    # may overwrite the start and end time. Currently these components and
-    # the overwritten recurrences are both considered. The overwritten
-    # recurrence should be ignored instead.
-
-    def getrruleset(child):
-        try:
-            first_dtstart = next(iter(child.getrruleset(addRDate=True)),
-                                 None)
-        except TypeError as e:
-            # TODO: The problem was fixed in VObject 0.9.5
-            raise VObjectBugException(
-                "failed to call getrruleset: %s" % e) from e
-        if first_dtstart is None:
-            # TODO: The problem was fixed in VObject 0.9.5
-            raise VObjectBugException(
-                "empty iterator from getrruleset")
+    def getrruleset(child, ignore=()):
         if (hasattr(child, "rrule") and
                 ";UNTIL=" not in child.rrule.value.upper() and
                 ";COUNT=" not in child.rrule.value.upper()):
-            if infinity_fn(_date_to_datetime(first_dtstart)):
-                return (), True
-        return child.getrruleset(addRDate=True), False
+            for dtstart in child.getrruleset(addRDate=True):
+                if dtstart in ignore:
+                    continue
+                if infinity_fn(_date_to_datetime(dtstart)):
+                    return (), True
+                break
+        return filter(lambda dtstart: dtstart not in ignore,
+                      child.getrruleset(addRDate=True)), False
+
+    def get_children(components):
+        main = None
+        recurrences = []
+        for comp in components:
+            if hasattr(comp, "recurrence_id") and comp.recurrence_id.value:
+                recurrences.append(comp.recurrence_id.value)
+                if comp.rruleset:
+                    # Prevent possible infinite loop
+                    raise ValueError("Overwritten recurrence with RRULESET")
+                yield comp, True, ()
+            else:
+                if main is not None:
+                    raise ValueError("Multiple main components")
+                main = comp
+        yield main, False, recurrences
 
     # Comments give the lines in the tables of the specification
     if child_name == "VEVENT":
-        for child in vobject_item.vevent_list:
+        for child, recurrence, recurrences in get_children(
+                vobject_item.vevent_list):
             # TODO: check if there's a timezone
             dtstart = child.dtstart.value
 
             if child.rruleset:
-                dtstarts, infinity = getrruleset(child)
+                dtstarts, infinity = getrruleset(child, recurrences)
                 if infinity:
                     return
             else:
@@ -338,30 +345,31 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
                 if dtend is not None:
                     # Line 1
                     dtend = dtstart + timedelta(seconds=original_duration)
-                    if range_fn(dtstart, dtend):
+                    if range_fn(dtstart, dtend, recurrence):
                         return
                 elif duration is not None:
                     if original_duration is None:
                         original_duration = duration.seconds
                     if duration.seconds > 0:
                         # Line 2
-                        if range_fn(dtstart, dtstart + duration):
+                        if range_fn(dtstart, dtstart + duration, recurrence):
                             return
                     else:
                         # Line 3
-                        if range_fn(dtstart, dtstart + SECOND):
+                        if range_fn(dtstart, dtstart + SECOND, recurrence):
                             return
                 elif dtstart_is_datetime:
                     # Line 4
-                    if range_fn(dtstart, dtstart + SECOND):
+                    if range_fn(dtstart, dtstart + SECOND, recurrence):
                         return
                 else:
                     # Line 5
-                    if range_fn(dtstart, dtstart + DAY):
+                    if range_fn(dtstart, dtstart + DAY, recurrence):
                         return
 
     elif child_name == "VTODO":
-        for child in vobject_item.vtodo_list:
+        for child, recurrence, recurrences in get_children(
+                vobject_item.vtodo_list):
             dtstart = getattr(child, "dtstart", None)
             duration = getattr(child, "duration", None)
             due = getattr(child, "due", None)
@@ -385,7 +393,7 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
                 created = _date_to_datetime(created.value)
 
             if child.rruleset:
-                reference_dates, infinity = getrruleset(child)
+                reference_dates, infinity = getrruleset(child, recurrences)
                 if infinity:
                     return
             else:
@@ -399,7 +407,7 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
                     reference_dates = (created,)
                 else:
                     # Line 8
-                    if range_fn(DATETIME_MIN, DATETIME_MAX):
+                    if range_fn(DATETIME_MIN, DATETIME_MAX, recurrence):
                         return
                     reference_dates = ()
 
@@ -409,56 +417,65 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
                 if dtstart is not None and duration is not None:
                     # Line 1
                     if range_fn(reference_date,
-                                reference_date + duration + SECOND):
+                                reference_date + duration + SECOND,
+                                recurrence):
                         return
                     if range_fn(reference_date + duration - SECOND,
-                                reference_date + duration + SECOND):
+                                reference_date + duration + SECOND,
+                                recurrence):
                         return
                 elif dtstart is not None and due is not None:
                     # Line 2
                     due = reference_date + timedelta(seconds=original_duration)
-                    if (range_fn(reference_date, due) or
+                    if (range_fn(reference_date, due, recurrence) or
                             range_fn(reference_date,
-                                     reference_date + SECOND) or
-                            range_fn(due - SECOND, due) or
-                            range_fn(due - SECOND, reference_date + SECOND)):
+                                     reference_date + SECOND, recurrence) or
+                            range_fn(due - SECOND, due, recurrence) or
+                            range_fn(due - SECOND, reference_date + SECOND,
+                                     recurrence)):
                         return
                 elif dtstart is not None:
-                    if range_fn(reference_date, reference_date + SECOND):
+                    if range_fn(reference_date, reference_date + SECOND,
+                                recurrence):
                         return
                 elif due is not None:
                     # Line 4
-                    if range_fn(reference_date - SECOND, reference_date):
+                    if range_fn(reference_date - SECOND, reference_date,
+                                recurrence):
                         return
                 elif completed is not None and created is not None:
                     # Line 5
                     completed = reference_date + timedelta(
                         seconds=original_duration)
                     if (range_fn(reference_date - SECOND,
-                                 reference_date + SECOND) or
-                            range_fn(completed - SECOND, completed + SECOND) or
+                                 reference_date + SECOND,
+                                 recurrence) or
+                            range_fn(completed - SECOND, completed + SECOND,
+                                     recurrence) or
                             range_fn(reference_date - SECOND,
-                                     reference_date + SECOND) or
-                            range_fn(completed - SECOND, completed + SECOND)):
+                                     reference_date + SECOND, recurrence) or
+                            range_fn(completed - SECOND, completed + SECOND,
+                                     recurrence)):
                         return
                 elif completed is not None:
                     # Line 6
                     if range_fn(reference_date - SECOND,
-                                reference_date + SECOND):
+                                reference_date + SECOND, recurrence):
                                 return
                 elif created is not None:
                     # Line 7
-                    if range_fn(reference_date, DATETIME_MAX):
+                    if range_fn(reference_date, DATETIME_MAX, recurrence):
                         return
 
     elif child_name == "VJOURNAL":
-        for child in vobject_item.vjournal_list:
+        for child, recurrence, recurrences in get_children(
+                vobject_item.vjournal_list):
             dtstart = getattr(child, "dtstart", None)
 
             if dtstart is not None:
                 dtstart = dtstart.value
                 if child.rruleset:
-                    dtstarts, infinity = getrruleset(child)
+                    dtstarts, infinity = getrruleset(child, recurrences)
                     if infinity:
                         return
                 else:
@@ -470,18 +487,18 @@ def _visit_time_ranges(vobject_item, child_name, range_fn, infinity_fn):
 
                     if dtstart_is_datetime:
                         # Line 1
-                        if range_fn(dtstart, dtstart + SECOND):
+                        if range_fn(dtstart, dtstart + SECOND, recurrence):
                             return
                     else:
                         # Line 2
-                        if range_fn(dtstart, dtstart + DAY):
+                        if range_fn(dtstart, dtstart + DAY, recurrence):
                             return
 
     elif isinstance(child, date):
-        if range_fn(child, child + DAY):
+        if range_fn(child, child + DAY, False):
             return
     elif isinstance(child, datetime):
-        if range_fn(child, child + SECOND):
+        if range_fn(child, child + SECOND, False):
             return
 
 
@@ -600,7 +617,7 @@ def find_tag_and_time_range(vobject_item):
         return (None, TIMESTAMP_MIN, TIMESTAMP_MAX)
     start = end = None
 
-    def range_fn(range_start, range_end):
+    def range_fn(range_start, range_end, recurrence):
         nonlocal start, end
         if start is None or range_start < start:
             start = range_start
@@ -913,10 +930,11 @@ def _propfind_response(base_prefix, path, item, props, user, write=False,
                 tag = ET.Element(_tag("D", "collection"))
                 element.append(tag)
             elif tag == _tag("D", "owner"):
-                if is_leaf and item.owner:
-                    element.text = "/%s/" % item.owner
-                else:
-                    is404 = True
+                # return empty elment, if no owner available (rfc3744-5.1)
+                if collection.owner:
+                    tag = ET.Element(_tag("D", "href"))
+                    tag.text = _href(base_prefix, "/%s/" % collection.owner)
+                    element.append(tag)
             elif tag == _tag("D", "displayname"):
                 if is_leaf:
                     element.text = item.get_meta("D:displayname") or item.path
@@ -935,7 +953,7 @@ def _propfind_response(base_prefix, path, item, props, user, write=False,
             else:
                 human_tag = _tag_from_clark(tag)
                 meta = item.get_meta(human_tag)
-                if meta:
+                if meta is not None:
                     element.text = meta
                 else:
                     is404 = True
@@ -1011,12 +1029,18 @@ def proppatch(base_prefix, path, xml_request, collection):
     href.text = _href(base_prefix, path)
     response.append(href)
 
-    for short_name in props_to_remove:
-        props_to_set[short_name] = ""
-    collection.set_meta(props_to_set)
-
-    for short_name in props_to_set:
+    new_props = collection.get_meta()
+    for short_name, value in props_to_set.items():
+        new_props[short_name] = value
         _add_propstat_to(response, short_name, 200)
+    for short_name in props_to_remove:
+        try:
+            del new_props[short_name]
+        except KeyError:
+            pass
+        _add_propstat_to(response, short_name, 200)
+    storage.check_and_sanitize_props(new_props)
+    collection.set_meta_all(new_props)
 
     return multistatus
 
@@ -1137,12 +1161,6 @@ def report(base_prefix, path, xml_request, collection):
                 if not all(match(item, filter_[0]) for filter_ in filters
                            if filter_):
                     continue
-            except VObjectBugException as e:
-                # HACK: Just return all items that can't be filtered because
-                # of bugs in VObject.
-                collection.logger.warning(
-                    "Failed to filter item %r from %r (Bug in VObject): %s",
-                    item.href, collection.path, e, exc_info=True)
             except Exception as e:
                 raise RuntimeError("Failed to filter item %r from %r: %s" %
                                    (item.href, collection.path, e)) from e
